import pandas as pd

def reduce_to_tracks(df, custom_agg=None):
    """

    Parameters
    ----------
    df pandas dataframe
    custom_agg: dict column : aggregation function or list of aggregation function

    Returns
    -------
    dataframe in which each row is a track, that correspond to rows of df aggregated by TrackHeadIndices for each Position, DatasetName
    """
    if 'TrackHeadIndices' not in df.columns:
        raise ValueError("TrackHeadIndices not in columns")
    # Default aggregation functions for tracking measurements
    agg_functions = {
        'Frame': ['min', 'max'],
        'TrackLength': 'first',
        'TrackObjectCount': 'first',
        'ParentTrackHeadIndices': 'first',
        'PositionIdx': 'first',
        'Time': 'first',
        'TrackErrorNext': 'max',  # True if any is True, False if all are False
        'TrackErrorPrev': 'max',  # True if any is True, False if all are False
        'NextDivisionFrame': 'first',
        'PreviousDivisionFrame' : 'first',
        'BacteriaLineage': 'first',
    }

    # Remove default aggregation columns that are not in the dataframe
    del_agg_cols = [agg_col for agg_col in agg_functions.keys() if agg_col not in df.columns]
    for del_agg_col in del_agg_cols:
        del agg_functions[del_agg_col]

    # Add custom aggregations if specified
    if custom_agg:
        agg_functions.update(custom_agg)
    else:
        custom_agg = {}

    # Add aggregation for other columns based on their data type, excluding custom_agg columns
    for col in df.columns:
        if col not in ['DatasetName', 'Position', 'Indices', 'TrackHeadIndices', 'Next', 'Prev', 'Idx'] and col not in agg_functions:
            if df[col].dtype == 'object':  # Assuming strings are of type 'object'
                agg_functions[col] = ['first', 'last']
            elif df[col].dtype == 'bool':
                agg_functions[col] = ['max', 'min']
            elif pd.api.types.is_numeric_dtype(df[col]):
                agg_functions[col] = ['mean', 'std']

    # Group by both TrackHeadIndices and Position, then aggregate
    group_cols = [col for col in ['DatasetName', 'Position', 'TrackHeadIndices'] if col in df.columns]
    track_df = df.groupby(group_cols).agg(agg_functions).reset_index()

    # Flatten the MultiIndex columns created by the aggregation
    track_df.columns = ['_'.join(col) if isinstance(col, tuple) else col for col in track_df.columns]

    # Rename columns to remove suffixes
    rename_cols =  { f"{c}_":c for c in group_cols }
    if 'Frame' in agg_functions:
        rename_cols['Frame_min'] = 'Frame'
        rename_cols['Frame_max'] = 'FrameLast'

    for c, agg in agg_functions.items():
        if isinstance(agg, str):
            rename_cols[f"{c}_{agg}"] = c
    track_df.rename(columns=rename_cols, inplace=True)

    return track_df



from os import listdir
from os.path import isfile, isdir, join, isabs
import json
import pandas as pd
from numpy import ndarray
import numpy as np
from ..selections import store_selection
from ..pandas import subset_by_DataFrame
import warnings

class Dataset():
    """Dataset object allow to facilitate manipulation of data generated by bacmman.

    Parameters
    ----------
    path : str
        path to the folder containing the configuration file
    data_path : str
        path to the folder containing the data files (optional, if None -> path is used)
    filter : str or callable
        filter on the name of the dataset (optional)

    Attributes
    ----------
    config_name : str
        dataset name
    object_class_names : list of str
        names of the object classes of the dataset. By default names are found in the configuration file, but can be renamed using the set_object_class_name method
    data : dict
        datatable for each object class
    path : str
        path to the folder containing the configuration file
    data_path : str
        (optional) path to the folder containing the data files. if None, path is used. If data_path is not absolute it is considered as relative to path
    filter: callable or str
        filter on dataset name. if str: test if filter is contained in dataset name

    """
    def __init__(self, path:str, data_path:str = None, filter=None, name:str=None, raise_error:bool=True):
        self.path = path
        if data_path is not None:
            self.data_path = data_path if isabs(data_path) else join(path, data_path)
        else:
            self.data_path = path
        self.config_name = get_dataset_name(path, filter)
        if self.config_name is not None: # inspect config file
            cf = join(path, self.config_name+"_config.json")
            with open(cf, errors='ignore') as f:
                try:
                    conf = json.load(f)
                    oc = conf["structures"]
                    if isinstance(oc, list):
                        self.object_class_names = [c["name"] for c in oc]  # new format
                    elif "list" in oc:
                        self.object_class_names = [c["name"] for c in oc["list"]]
                    else:
                        raise IOError(f"Invalid configuration format: could not find object class names in {oc}")
                    # ensure unicity of names
                    seen = set()
                    self.object_class_names = [_increment_name(x, seen) for x in self.object_class_names]
                except Exception as e:
                    print(f"Error trying to load configuration file: {f}")
                    raise e
        else:
            self.object_class_names = None
            if raise_error:
                raise IOError(f"Invalid dataset directory : {path}")
        self.name = self.config_name if name is None else name
        self.data = {}
        self.selections = None

    def set_object_class_name(self, old_name, new_name:str):
        """Modifies the name of an object class.

        Parameters
        ----------
        old_name : str or int
            either name or index of the old object class to modify
        new_name : str
            new name of the object class

        """
        old_name = self._get_object_class_index(old_name)
        assert old_name < len(self.object_class_names), f"invalid object class idx, should be <{len(self.object_class_names)}"
        if self.object_class_names[old_name]!=new_name:
            assert new_name not in self.object_class_names, "invalid new object_class name: already exists"
            self.object_class_names[old_name] = new_name

    def _get_object_class_name(self, object_class):
        if not isinstance(object_class, str):
            assert object_class<len(self.object_class_names), f"invalid object_class index, should be < {len(self.object_class_names)}"
            return self.object_class_names[object_class]
        else:
            return object_class

    def _get_object_class_index(self, object_class):
        if isinstance(object_class, str):
            assert object_class in self.object_class_names, f"invalid object_class name, should be in {self.object_class_names}"
            return self.object_class_names.index(object_class)
        else:
            return object_class

    def _get_data_file_path(self, object_class):
        return join(self.data_path, f"{self.config_name}_{self._get_object_class_index(object_class)}.csv")

    def _open_data(self, object_class, add_dataset_name_column=False, **kwargs):
        default_dtype = {'Position': str, 'PositionIdx':np.int16, 'Frame':np.int16, "Idx":np.int16,
                         'Indices':str, 'TrackHeadIndices':str, 'ParentTrackHeadIndices':str, 'Prev':str,
                         'Next':str, 'TrackErrorNext':bool, 'TrackErrorPrev':bool}
        if 'dtype' in kwargs and kwargs['dtype'] is not None:
            kwargs['dtype'].update(default_dtype)
        else:
            kwargs['dtype'] = default_dtype
        data = pd.read_csv(self._get_data_file_path(object_class), sep=';', **kwargs) #
        if add_dataset_name_column:
            data["DatasetName"] = self.name
        return data

    def get_data(self, object_class, selection=None, copy:bool=True, cache:bool=True, **kwargs):
        """Open measurement table of an object class as pandas dataframe.
           the file is supposed to be located in the dataset path.
           Note that when calling this method, the dataframe is stored in the dataset object. call _open_data to open the dataframe without storing it.

        Parameters
        ----------
        object_class : str or int
            either name or index of the object class
        selection: str or dataframe
            str : name of a selection contained in this dataset
            dataframe : dataframe that contains columns
        copy: bool
            only used if selection is None. returned a copy of the dataframe to avoid modifying the cached one
        cache: bool
            cache the dataframe
        Returns
        -------
        pandas dataframe
            Description of returned object.

        """
        object_class_name = self._get_object_class_name(object_class)
        object_class_idx = self._get_object_class_index(object_class)
        if object_class_name not in self.data:
            data = self._open_data(object_class, **kwargs)
            if cache:
                self.data[object_class_name] = data
        else:
            data = self.data[object_class_name]
        if selection is not None:
            if isinstance(selection, str):
                selections = self.get_selections()
                if selections is None:
                    raise ValueError(f"No selections found in dataset {self}")
                selection = selections[(selections.ObjectClassIdx == object_class_idx) & (selections.SelectionName == selection)]
                if selection.shape[0]==0:
                    warnings.warn(f"Selection {selection} not found for object class: {object_class}")
            elif isinstance(selection, (list, tuple)):
                selections = self.get_selections()
                selections = selections[selections.ObjectClassIdx == object_class_idx]
                all_selections = [selections[selections.SelectionName == s] if isinstance(s, str) else s for s in selection]
                for s, sname in zip(all_selections, selection):
                    if isinstance(sname, str) and s.shape[0]==0:
                        warnings.warn(f"Selection: {sname} not found for object class: {object_class}")
                selection = pd.concat(all_selections)
                selection = selection.drop_duplicates(["Position", "Indices"])
            return subset_by_DataFrame(data, selection, on=["Position", "Indices"])
        else:
            return data.copy() if copy else data

    def _get_selections_file_path(self):
        return join(self.data_path, f"{self.config_name}_Selections.csv")

    def _open_selections(self, add_dataset_name_column=False, **kwargs):
        default_dtype = {'Position': 'str', 'PositionIdx':np.int16, 'ObjectClassIdx':np.int16, 'Indices':'str', 'Frame':np.int16, 'SelectionName':'str'}
        if 'dtype' in kwargs and kwargs['dtype'] is not None:
            kwargs['dtype'].update(default_dtype)
        else:
            kwargs['dtype'] = default_dtype
        fp = self._get_selections_file_path()
        if not isfile(fp):
            columns = ['Position', 'PositionIdx', 'ObjectClassIdx', 'Indices', 'Frame', 'SelectionName']
            if add_dataset_name_column:
                columns.append('DatasetName')
            return pd.DataFrame(columns=columns)
        else :
            data = pd.read_csv(fp, sep=';', **kwargs) #
        if add_dataset_name_column:
            data["DatasetName"] = self.name
        return data

    def append_selections_to_data(self, selections=None):
        """For each selection, adds a boolean column to the dataFrame with the name of the selection, with True where the object defined by (Position, Indices) is included in the selection. Object class is infered from the selection name.

        Parameters
        ----------
        selections : list / tuple / ndarray of str, or str
            Selection names to add. If None, all selections will be added

        Returns None

        """
        all_selections = self.get_selections()
        if all_selections is None:
            warnings.warn(f"No selections found in dataset: {self.name}")
            return
        if selections is None:
            selections = all_selections["SelectionName"].unique()
            print(type(selections))
        if isinstance(selections, str):
            selections = [selections]
        if not isinstance(selections, (tuple, list, ndarray)):
            raise ValueError("selections should be a str, tuple or list of str, or ndarray")
        for sel_name in selections:
            sel = all_selections[all_selections["SelectionName"] == sel_name]
            if sel.shape[0] == 0:
                warnings.warn(f"Selection: {sel_name} not found in exported selections")
                continue
            oc_idx = sel["ObjectClassIdx"].iloc[0]
            data = self.get_data(oc_idx)
            if data is None:
                warnings.warn(f"No data found for object class: {oc_idx}, skipping selection: {sel_name}")
            else:
                merge = data.filter(["Position", "Indices"], axis=1)
                merge['copy_index'] = data.index
                merge = merge.merge(sel[["Position", "Indices"]], how="inner", copy=False)["copy_index"]
                data[sel_name] = False
                data.loc[merge, sel_name] = True

    def get_selections(self, **kwargs):
        if self.selections is None:
            self.selections = self._open_selections(**kwargs)
        return self.selections

    def store_selection(self, selection, object_class, name:str, **kwargs):
        """Save a selection (subset of objects) to BACMMAN software.

        Parameters
        ----------
        object_class : str or int
            either name or index of the object class
        selection : pandas dataframe
            dataframe containing indices and positions of objects to include in the selection
        name : name of the selection
        **kwargs : dict
            extra arguments passed to save_selection method, such as port, python_proxy_port, address

        """
        object_class_idx = self._get_object_class_index(object_class)
        store_selection(selection, self.config_name, objectClassIdx=object_class_idx, selectionName=name, dsPath=self.path, **kwargs)

    def __getitem__(self, item):
        return self.get_data(item)

    def __str__(self):
        return f"{self.name} oc={self.object_class_names} path={self.path}"+(f"data path={self.data_path}" if self.data_path!=self.path else "")

class DatasetList(Dataset):
    def __init__(self, dataset_list:list = None, path:str = None, data_path:str = None, config_path:str=None, filter = None, object_class_name_mapping:dict = None):
        if dataset_list is not None:
            self.datasets = {d.name:d for d in dataset_list}
        elif path is not None:
            if config_path is not None:
                assert data_path is None, "data_path is not taken into account when config_path is provided"
                dataset_list = [Dataset(path=config_path, data_path=join(path, f), filter=filter, name = f, raise_error=False) for f
                                in listdir(path) if isdir(join(path, f))]
                self.datasets = {d.name: d for d in dataset_list if d.name is not None}
            else:
                if data_path is not None:
                    assert not isabs(data_path), "data_path must be relative"
                dataset_list = [Dataset(path=join(path, f), data_path=data_path, filter=filter, raise_error=False) for f in listdir(path) if isdir(join(path, f))]
                self.datasets = {d.name:d for d in dataset_list if d.name is not None}
            if object_class_name_mapping is not None:
                for d in self.datasets.values():
                    for i, n in enumerate(d.object_class_names):
                        if n in object_class_name_mapping and object_class_name_mapping[n] not in d.object_class_names:
                            d.object_class_names[i] = object_class_name_mapping[n]

        assert len(self.datasets)>1, "no datasets where found"
        # object_class names are the common object_class names between all object_class names
        self.object_class_names = None
        for d in self.datasets.values():
            if self.object_class_names is None:
                self.object_class_names = d.object_class_names
            else:
                inter = set(self.object_class_names).intersection(d.object_class_names)
                self.object_class_names = [n for n in self.object_class_names if n in inter]
        assert self.object_class_names is not None, "no object classes found"
        assert len(self.object_class_names)>0, f"no object_class names in common between datasets : {[str(d) for d in self.datasets.values()]}"
        self.data = {}

    def __getitem__(self, item):
        return self.datasets[item]

    def set_object_class_name(self, old_object_class, new_object_class_name):
        old_object_class_idx = self._get_object_class_index(old_object_class)
        old_object_class_name = self._get_object_class_name(old_object_class)
        assert old_object_class_idx < len(self.object_class_names), f"invalid object_class idx, should be <{len(self.object_class_names)}"
        self.object_class_names[old_object_class_idx] = new_object_class_name
        for d in self.datasets.values():
            d.set_object_class_name(old_object_class_name, new_object_class_name)

    def _open_data(self, object_class, **kwargs):
        object_class = self._get_object_class_name(object_class)
        return pd.concat([d.get_data(object_class, add_dataset_name_column=True, copy=False, cache=False, **kwargs) for d in self.datasets.values()]) # do not cache both in each dataset and datasetList

    def store_selection(self, selection, object_class, name:str, dataset_column="DatasetName", **kwargs):
        object_class_name = self._get_object_class_name(object_class)
        for ds_name, sel in selection.groupby(dataset_column):
            if ds_name not in self.datasets:
                print(f"Error: dataset {ds_name} not in datasets: {str(self)}")
            else:
                self.datasets[ds_name].store_selection(sel, object_class_name, name, **kwargs)

    def __str__(self):
        return f"{[n for n in self.datasets.keys()]} oc={self.object_class_names}"

# util functions
def get_dataset_name(path, filter):
    # must contain a file ending by _config.json
    if isinstance(filter, str):
        _filter = lambda file_name:filter in file_name
    else:
        _filter = filter
    if isdir(path):
        for f in listdir(path):
            if f.endswith("_config.json") and (filter is None or _filter(f[:-12])) and isfile(join(path, f)):
                return f[:-12]
    return None

def _increment_name(name, seen):
    if name in seen:
        new_name = name+"_1"
        inc = 2
        while new_name in seen:
            new_name = f"{name}_{inc}"
            inc += 1
        name = new_name
    seen.add(name)
    return name




import numpy as np


def interpolateX(x1, y1, x2, y2):
    x_tot=np.concatenate((x1,x2),axis=0)
    x_tot=np.sort(x_tot)
    y1_tot=np.interp(x_tot,x1,y1)
    y2_tot=np.interp(x_tot,x2,y2)
    return x_tot, y1_tot, y2_tot

def compute_aera_under_min(x, y1, y2):
    y_min=np.minimum(y1,y2);
    return np.trapz(y_min,x)

def interpolated_intercepts(x, y1, y2):
    """Find the intercepts of two curves, given by the same x data"""

    def intercept(point1, point2, point3, point4):
        """find the intersection between two lines
        the first line is defined by the line between point1 and point2
        the first line is defined by the line between point3 and point4
        each point is an (x,y) tuple.

        So, for example, you can find the intersection between
        intercept((0,0), (1,1), (0,1), (1,0)) = (0.5, 0.5)

        Returns: the intercept, in (x,y) format
        """

        def line(p1, p2):
            A = (p1[1] - p2[1])
            B = (p2[0] - p1[0])
            C = (p1[0]*p2[1] - p2[0]*p1[1])
            return A, B, -C

        def intersection(L1, L2):
            D  = L1[0] * L2[1] - L1[1] * L2[0]
            Dx = L1[2] * L2[1] - L1[1] * L2[2]
            Dy = L1[0] * L2[2] - L1[2] * L2[0]

            x = Dx / D
            y = Dy / D
            return x,y

        L1 = line([point1[0],point1[1]], [point2[0],point2[1]])
        L2 = line([point3[0],point3[1]], [point4[0],point4[1]])

        R = intersection(L1, L2)

        return R

    idxs = np.argwhere(np.diff(np.sign(y1 - y2)) != 0)

    xcs = []
    ycs = []

    for idx in idxs:
        xc, yc = intercept((x[idx], y1[idx]),((x[idx+1], y1[idx+1])), ((x[idx], y2[idx])), ((x[idx+1], y2[idx+1])))
        xcs.append(xc[0])
        ycs.append(yc[0])
    return xcs, ycs



from py4j.java_gateway import JavaGateway, GatewayParameters # requires py4j
from py4j.java_collections import ListConverter
from py4j.protocol import Py4JNetworkError
import json
import os

def store_selection(df, dsName:str, objectClassIdx:int, selectionName:str, dsPath:str=None, showObjects:bool=False, showTracks:bool=False, openSelection:bool=False, objectClassIdxDisplay:int=-1, interactiveObjectClassIdx:int=-1, port:int=None, python_proxy_port:int=None, address:str=None, gateway_parameters={}, **kwargs):
    """Stores a selection to bacmman using python gateway (py4j). Bacmman must be running with an active python gateway server.

    Parameters
    ----------
    df : pandas DataFrame
        each line of the DataFrame is one element of the selection, defined by columns Indices & Position
    dsName : str
        bacmman dataset name to store the selection to.
    dsPath : str
        path to the folder containing the bacmman dataset. Can be omitted if BACMMAN is open and dsName is a relative name to the working directory currently set in BACMMAN
    objectClassIdx : int
        index of the object class of the elements of the selection in the bacmman dataset
    selectionName : str
        name of the selection
    showObjects : bool
        whether contours of objects should be shown
    showTracks : bool
        whether track links of objects should be shown
    openSelection : bool
        whether the first kymograph of the selection should be open
    objectClassIdxDisplay : int
        if openSelection is true, object class idx of the opened kymograph
    interactiveObjectClassIdx : int
        if openSelection is true, interactive object class idx
    python_proxy_port : int
        python port of the java gateway
    """
    if port is None:
        port = int(os.getenv("PYBACMMAN_PORT", 25335))
    if python_proxy_port is None:
        python_proxy_port = int(os.getenv("PYBACMMAN_PYPROXYPORT", 25334))
    if address is None:
        address = os.getenv("PYBACMMAN_ADDRESS", '127.0.0.1')
    # fix path if called from from docker
    container_dir = os.getenv("BACMMAN_CONTAINER_DIR")
    wd = os.getenv("BACMMAN_WD")
    if container_dir is not None and wd is not None and container_dir in dsPath:
        dsPath = dsPath.replace(container_dir, wd)
    gateway = JavaGateway(python_proxy_port=python_proxy_port, gateway_parameters=GatewayParameters(address=address, port=port, **gateway_parameters))
    try:
        idx = ListConverter().convert(df.Indices.tolist(), gateway._gateway_client)
        pos = ListConverter().convert(df.Position.tolist(), gateway._gateway_client)
        gateway.saveCurrentSelection(dsName, dsPath, objectClassIdx, selectionName, idx, pos, showObjects, showTracks, openSelection, False, objectClassIdxDisplay, interactiveObjectClassIdx)
    except Py4JNetworkError as err:
        # try to fallback to saveSelectionFile method if dsPath is provided
        if dsPath is None:
            print("Could not connect, is BACMMAN started? Check that Python Gateway parameters match (BACMMAN menu MISC>Python Gateway)")
            print(err)
            print("provide path to dataset as dsPath in kwargs to save selection as a file")
            print(err)
        else:
            print(f"Could not connect through python gateway, saving selection as a file to {dsPath}")
            store_selection_file(df, dsPath=dsPath, objectClassIdx=objectClassIdx, selectionName=selectionName, **kwargs)

def store_selection_file(df, dsPath:str, objectClassIdx:int, selectionName:str, indexCol:str= "Indices", positionCol:str= "Position"):
    data = {
        "name": selectionName,
        "objectClassIdx": objectClassIdx,
        "objects": {position : list(g[indexCol]) for position, g in df.groupby(positionCol)}
    }
    json_object = json.dumps(data)
    sel_path = os.path.join(dsPath, "Output", "Selections")
    assert os.path.exists(dsPath), f"dataset path not found: {dsPath} cannot save selection"
    if not os.path.exists(sel_path): # check path and create if necessary
        os.makedirs(sel_path, exist_ok=True)
        if not os.path.exists(sel_path):
            print(f"could not create selection dir: {sel_path}, cannot save selection")
    file_path = os.path.join(sel_path, selectionName + ".json")
    with open(file_path, "w") as outfile:
        outfile.write(json_object)



# bacmman indices (barcode) manipulation
def get_parent(indices):
    split = indices.split('-')[:-1]
    if len(split)==1:
        return int(split[0])
    else:
        return '-'.join(split)
get_frame = lambda indices : int(indices.split('-')[0])
def get_previous(currentIndices):
    spl = currentIndices.split('-')
    spl[0] = str(int(spl[0])-1)
    return '-'.join(spl)
def get_next(currentIndices):
    spl = currentIndices.split('-')
    spl[0] = str(int(spl[0])+1)
    return '-'.join(spl)
def set_frame(indices, newFrame):
    spl = indices.split('-')
    spl[0] = str(newFrame)
    return '-'.join(spl)



import pandas as pd
from pandas import merge
from .indices_utils import get_next, get_previous
import math
import matplotlib.pyplot as plt

def subset_by_DataFrame(df, dfSubset, on, sub_on=None, keepCols=[], remove:bool=False):
    """Return rows of dataframe df that are present in dfSubset

    Parameters
    ----------
    df : DataFrame
        DataFrame to subset
    dfSubset : DataFrame
        DataFrame used to define the subset
    on : list of Strings
        Column names to join on. These must be found in both DataFrames if sub_on is None.
    sub_on : list of Strings
        Column names to join on in the dfSubset DataFrame.
    keepCols : list of Strings
        columns of dfSubset that should be kept in resulting DataFrame
    remove : boolean
        if true, rows of df matching with dfSubset will be removed.
        if false, only rows of df matching with a row of dfSubset will be kept
    Returns
    -------
    DataFrame
        subset of df

    """
    if sub_on is None:
        sub_on=on
        rem_cols=[]
    else:
        assert len(sub_on)==len(on), "sub_on should have same length as on"
        rem_cols = [c+'_y' for c in sub_on if not c in on]
    if len(keepCols)==0:
        subcols = sub_on
    else:
        subcols = sub_on + keepCols
    dfSubset = dfSubset[subcols].drop_duplicates()
    if not remove:
        res = merge(df, dfSubset, how='inner', left_on=on, right_on=sub_on, suffixes=('', '_y'))
    else:
        res = merge(df, dfSubset, how='left', left_on=on, right_on=sub_on, suffixes=('', '_y'), indicator=True)
        res = res.query('_merge != "both"')
        rem_cols.append("_merge")
    if len(rem_cols)>0:
        res.drop(rem_cols, 1, inplace=True)
    return res

def map_columns(df, df2, columns, on, on2=None):
    """Return mapped columns from df2 with the index of df using on (resp. on2) columns to match rows

    Parameters
    ----------
    df : DataFrame
        DataFrame to subset
    df2 : DataFrame
        DataFrame used to define the subset
    columns : String or list of Strings
        column(s) of df2 to be returned
    on : list of Strings
        Column names to map df and df2. These must be found in both DataFrames if on2 is None.
    on2 : list of Strings
        Column names in the df2 DataFrame to be mapped to on columns in df.


    Returns
    -------
    DataFrame
        list of columns with rows

    """
    if on2 is None:
        on2=on
    if not isinstance(on2, list):
        on2 = [on2]
    if not isinstance(on, list):
        on = [on]
    assert len(on2)==len(on), "on2 should have same length as on"
    if not isinstance(columns, list):
        columns = [columns]
    subcols = on2 + columns
    df2 = df2[subcols].drop_duplicates()
    res = pd.merge(df[on], df2, how='left', left_on=on, right_on=on2)
    res.index = df.index
    if len(columns)==1:
        return res.loc[:,columns[0]]
    else:
        return [res.loc[:,c] for c in columns]

def merge_dataframes(df, df2, on, on2=None, columns_from_df2=None):
    """
    Merge two DataFrames based on specified columns and retain the index of the first DataFrame.

    Parameters:
    df (pd.DataFrame): The first DataFrame.
    df2 (pd.DataFrame): The second DataFrame.
    on (str or list): The column(s) in df to merge on.
    on2 (str or list, optional): The column(s) in df2 to merge on. Defaults to on if not provided.
    columns_from_df2 (list, optional): List of columns from df2 to include in the merged DataFrame.
                                       If None, all columns from df2 will be included.

    Returns:
    pd.DataFrame: The merged DataFrame with the index of df.
    """
    if on2 is None:
        on2 = on
    if not isinstance(on2, list):
        on2 = [on2]
    if not isinstance(on, list):
        on = [on]
    assert len(on2) == len(on), "on2 should have the same length as on"
    if columns_from_df2 is None:
        columns_from_df2 = df2.columns.tolist()
    for on2_el in on2:
        if on2_el not in columns_from_df2:
            columns_from_df2.append(on2_el)
    merged_df = pd.merge(df, df2[columns_from_df2], left_on=on, right_on=on2, how='left', suffixes=(None, '_df2'))
    cols_to_drop = [f"{col}_df2" for col in on2 if col not in on]
    if len(cols_to_drop)>0:
        merged_df.drop(columns=cols_to_drop, inplace=True)
    merged_df.index = df.index

    return merged_df

def set_selection_edges(dfSelection, left=True):
    """Create a boolean column that defines the edge of the selection (in terms of Frames). If left=True: value is True when there is no previous object. If left = false, value is True when there is no next object

    Parameters
    ----------
    dfSelection : DataFrame
        dataframe with a column SelectionName, Position & Indices
    left : boolean
        see summary

    Returns
    -------
    void
        inplace modification of dfSelection

    """
    colName = "LeftEdge" if left else "RightEdge"
    fun = get_next if left else get_previous
    dfNeigh = pd.DataFrame({"SelectionName":dfSelection.SelectionName, "Position":dfSelection.Position, "Indices":dfSelection.Indices.apply(fun)})
    leftEdges = pd.concat([dfSelection[["SelectionName", "Position", "Indices"]], dfNeigh, dfNeigh]).drop_duplicates(keep=False)
    dfSelection[colName] = False
    dfSelection.loc[leftEdges.index, colName] = True

def group_plot(groupedData, plotFun, xlabel=None, ylabel=None, ncols=4, figsize=(12, 4)):
    """Short summary.

    Parameters
    ----------
    groupedData : grouped dataframe
        Description of parameter `groupedData`.
    plotFun : function
        inputs : group and pyplot axe and plot graph on the axe
    xlabel : type
        Description of parameter `xlabel`.
    ylabel : type
        Description of parameter `ylabel`.
    ncols : type
        Description of parameter `ncols`.
    figsize : type
        Description of parameter `figsize`.

    Returns
    -------
    groupPlot(groupedData, plotFun, xlabel=None, ylabel=None, ncols=4,
        Description of returned object.

    """
    ncols=min(ncols, groupedData.ngroups)
    nrows = int(math.ceil(groupedData.ngroups/ncols))
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, sharex=True, sharey=True)
    if ncols>1 or nrows>1:
        axflat =  axes.flatten()
    else:
        axflat = [axes]
    for (key, ax) in zip(groupedData.groups.keys(),axflat):
        data = groupedData.get_group(key)
        plotFun(data, ax)
        ax.set_title(key)
    ax.legend()
    if xlabel:
        fig.text(0.5, 0.02, xlabel, ha='center')
    if ylabel:
        fig.text(0.08, 0.5, ylabel, va='center', rotation='vertical')
    return fig, axes



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PyBacmman",
    version="0.6.4",
    author="Jean Ollion",
    author_email="jean.ollion@polytechnique.org",
    description="Utilities for analysis of data generated from bacmman software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jeanollion/PyBacmman.git",
    download_url = 'https://github.com/jeanollion/PyBacmman/archive/v0.6.4.tar.gz',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: OS Independent",
    ],
    keywords = ['bacmman', 'pandas', 'data analysis'],
    python_requires='>=3',
    install_requires=['py4j', 'pandas']

)



import shutil
import subprocess
import os
from os import listdir
from os.path import isfile, join

def run_bacmman_task(bacmmanFolder:str, taskFile:str, logFile:str=None, progressCallback=None, verbose=1):
    """Helper method to execute a bacmman task command, display its progress and log messages

    Parameters
    ----------
    bacmmanFolder : str
        directory containing of bacmman-headless-<VERSION>.jar and dependencies. e.g. compile bacmman with the command: mvn clean dependency:copy-dependencies package -DskipTests
    taskFile : str
        path of the .json task file to execute
    logFile : str
        None or path of a file where all log messages will be append
    progressCallback : executable
        callback to display progress of the task. Called with the progress as an int in range [0,100]. use getJNBProgressBar for jupyter notebooks
    verbose : int in range [0,2]
        verbose level. if >1 all log message are displayed, if ==1 only main messages are displayed, if 0 none are displayed

    Returns
    -------
    type
        result of runCommand method
    """

    # grab bacmman-headless file (depends on version)
    getjar = lambda folder : [join(folder,f) for f in listdir(folder) if f.startswith("bacmman-headless-") and f.endswith(".jar")]
    bheadlessjar = getjar(bacmmanFolder)
    if len(bheadlessjar)==0:
        if os.path.isdir(join(bacmmanFolder, "target")):
            bacmmanFolder = join(bacmmanFolder, "target")
            bheadlessjar = getjar(bacmmanFolder)
        elif os.path.isdir(join(bacmmanFolder, "bacmman-headless", "target")):
            bacmmanFolder = join(bacmmanFolder, "bacmman-headless", "target")
            bheadlessjar = getjar(bacmmanFolder)
        elif os.path.isdir(join(bacmmanFolder, "bacmman", "bacmman-headless", "target")):
            bacmmanFolder =join(bacmmanFolder, "bacmman", "bacmman-headless", "target")
            bheadlessjar = getjar(bacmmanFolder)

    if len(bheadlessjar)==0:
        raise ValueError("bacmman-headless .jar file not found in dir:", bacmmanFolder)
    elif len(bheadlessjar)>1:
        raise ValueError("{} bacmman-headless .jar file found in dir: {}".format(len(bheadlessjar), bacmmanFolder))
    dep_folder = join(bacmmanFolder, "dependency")
    if not os.path.isdir(dep_folder):
        raise ValueError("dependency folder not found in dir:", bacmmanFolder)
    if not os.path.isfile(taskFile):
        raise ValueError("task file {} not found".format(taskFile))
    if os.name == 'nt': # windows
        cmd = "java -cp {}\*;{} bacmman.ui.ProcessTasks {}{}".format(dep_folder, bheadlessjar[0], taskFile, " "+logFile if logFile else "")
    else:
        cmd = "java -cp {}/*:{} bacmman.ui.ProcessTasks {}{}".format(dep_folder, bheadlessjar[0], taskFile, " "+logFile if logFile else "")
    print("Command : ", cmd)
    return run_command(cmd, progressCallback, verbose)

def run_command(cmd:str, progressCallback=None, verbose = 1):
    """Helper method to execute a (bacmman task) command, display its progress and log messages

    Parameters
    ----------
    cmd : str
        shell command
    progressCallback : executable
        callback to display progress of the task. Called with the progress as an int in range [0,100]. use getJNBProgressBar for jupyter notebooks
    verbose : int in range [0,2]
        verbose level. if >1 all log message are displayed, if ==1 only main messages are displayed, if 0 none are displayed

    Returns
    -------
    type
        tuple (error, exit_code)
        error : whether there was errors during the execution of the task by BACMMAN
        exit_code: exit code of the execution of the shell command

    """
    import shlex
    errorMessages = (">Error", ">Invalid Task Error")
    noErrorMessages=(">Errors: 0")
    logMessages = (">Run task: ", ">All jobs finished")
    with subprocess.Popen(shlex.split(cmd), stdout=subprocess.PIPE, bufsize=1, universal_newlines=True) as p:
        error = False
        if verbose>0:
            if progressCallback is not None:
                progressCallback(0)
            for line in p.stdout:
                if line.startswith(">Progress: "):
                    if progressCallback is not None:
                        progressCallback(int(line.split(" ")[1][:-2]))
                elif line.startswith(errorMessages):
                    if not line.startswith(noErrorMessages):
                        error = True
                        print(line, end='', flush=True)
                elif verbose>1 or line.startswith(logMessages):
                    print(line, end='', flush=True)
        exit_code = p.poll()
    return error, exit_code

def get_jnb_progress_bar():
    '''
        return a progress bar for jupyter-lab / jupyter notebook, to be used with the runCommand method
        requires ipywidgets
    '''
    from ipywidgets import IntProgress
    from IPython.display import display
    pb = IntProgress(min=0, max=100)
    display(pb)
    def callback(value):
        pb.value = value
    return callback
